/*
	parse.c
 
    This file is part of munpak.
 
    Routines to actually parse the MPAKs

    Copyright (c) 2002 @stake, Inc.
*/

#include <stdio.h>
#include <windows.h>
#include <string.h>
#include <time.h>
#include "mobitex.h"
#include "munpak.h"

void parse (unsigned long man, FILE* fp)
{
    MPAK    mpak[1];
    byte    instr[100], block[18], strtmp[10];
    int     line, badcrc, disp = 0;
    unsigned long tmp;
    DONELIST *dl = NULL, *entry;
    
    if (man)
        fprintf(stdout, "\nOnly parsing Mobitex Access Number (MAN): %d", man);

    while (1)
    {
        if (fseek(fp, 0L, SEEK_SET)) // go to beginning of file
        {
            fprintf(stderr, "\nFile seek failed.\n");
            exit(1);
        }

        while (!feof(fp))
        {
            line = 0;
            badcrc = 0;

            while (fgets(instr, 100, fp))
            {
                if (ferror(fp))
                {
                    fprintf(stdout, "\nRead error.\n");
                    exit(1);
                }

                if(strstr(instr, "BAD CRC") && !badcrc)
                    badcrc = 1; // set flag to print warning later

                instr[36] = '\0'; // first 36 characters of line (crc removed)
	            if (atoc (instr, block)) // convert from ascii to raw 18-byte data block
                {
                    fprintf(stderr, "\nCharacter conversion error.\n");
                    exit(1);
                }

                // stuff into MPAK structure
                if (line == 0 || line < mpak->Rosi.blocks)
                {
                    if (instr[0] != 0x0A) // ignore unused lines
                    {
                        memcpy((struct MPAK*) ((byte*) mpak + (line * 18)), block, 18);

                        if (mpak->Rosi.blocks > 1)
                            ++line;
                        else
                            if (badcrc) badcrc = 0; // if single line message has bad crc, clear flag
                    }
                }
                else
                    break; // done reading message
            }

            sprintf(strtmp, "%.2X%.2X%.2X", mpak->Rosi.man[0], mpak->Rosi.man[1], mpak->Rosi.man[2]);
            tmp = strtoul(strtmp, NULL, 16);
            if (man == 0 || man == tmp)
            {
                fprintf(stdout, "\n\n===========================================================================\n");

                if (badcrc)
                    fprintf(stdout, "\n\nMPAK contains bad CRC.\n");

        	    while (dl != NULL) 
                {
                    if ((memcmp(&(dl->man), mpak->Rosi.man, 3) == 0) &&
                        (memcmp(&(dl->packet), &mpak->Header.packet, 1) == 0))
		    
                    {
                        display (mpak, 1);
                        disp = 1;
                        break;
                    }
                    else
                        disp = 0;

		            dl = dl->next;
	            }
                if (!disp)
                    display (mpak, 0); // if this is the first packet (e.g., MAN/Packet ID not in the list)
            }

            // whole MPAK is processed, so print results
            // Keep track of which MPAKs have been parsed,
            // and loop to find next MPAK in message, if it exists
            // MAN and Packet ID stay constant, Sequence Number increases
            entry = (DONELIST *) malloc(sizeof(DONELIST));
		    memcpy (&(entry->man), mpak->Rosi.man, 3); 
            memcpy (&(entry->sequence), &mpak->Rosi.sequence, 1); 
            memcpy (&(entry->packet), &mpak->Header.packet, 1); 
		    entry->next = dl;
		    dl = entry;
        }

        if (feof(fp))
        {
            fprintf(stdout, "\n");
            freelist(dl);
	        if (fp) fclose(fp);
            exit(0);
        }
    }
}


void display (MPAK *mpak, int packet)
{
    byte strtmp[522];
    int tmp, s;

    if (!packet) // if it is the first packet in the message, print all header info
    {
        MTYPE   type; // message type

        fprintf(stdout, "\nRadio Oriented Synchronous Information (ROSI) Header");
        fprintf(stdout, "\n----------------------------------------------------");

        sprintf(strtmp, "%.2X%.2X%.2X", mpak->Rosi.man[0], mpak->Rosi.man[1], mpak->Rosi.man[2]);
        tmp = strtoul(strtmp, NULL, 16);
        fprintf(stdout, "\nMobitex Access Number (MAN): %d", tmp);
        fprintf(stdout, "\nFrame ID: %d", mpak->Rosi.frame);
        fprintf(stdout, "\nSequence Number: %d", mpak->Rosi.sequence);
        fprintf(stdout, "\nData Blocks: %d", mpak->Rosi.blocks);

        fprintf(stdout, "\n\nMobitex Packet (MPAK) Header");
        fprintf(stdout, "\n----------------------------");

        sprintf(strtmp, "%.2X%.2X%.2X", mpak->Header.sender[0], mpak->Header.sender[1], mpak->Header.sender[2]);
        tmp = strtoul(strtmp, NULL, 16);
        fprintf(stdout, "\nSender MAN: %d", tmp);

        sprintf(strtmp, "%.2X%.2X%.2X", mpak->Header.dest[0], mpak->Header.dest[1], mpak->Header.dest[2]);
        tmp = strtoul(strtmp, NULL, 16);
        fprintf(stdout, "\nAddressee MAN: %d", tmp);

        // if source address is same as ROSI MAN, device is transmitting: show flags
        // else device is receiving: show traffic state
        fprintf(stdout, "\nFlags: ");
        if (memcmp(mpak->Rosi.man, mpak->Header.sender, 3) == 0)
        {
            if (mpak->Header.flags & MAILBOX_FLAG_MASK)
                fprintf(stdout, "Store & forward mail; ");

            if (mpak->Header.flags & POSACK_FLAG_MASK)
                fprintf(stdout, "Delivery confirmation; ");

            if (mpak->Header.flags & SENDLIST_FLAG_MASK)
                fprintf(stdout, "Use address list");

            if (!(mpak->Header.flags & 0x7))
                fprintf(stdout, "None");
        }
        else
            fprintf(stdout, "N/A");

        fprintf(stdout, "\nTraffic State: ");
        if (memcmp(mpak->Rosi.man, mpak->Header.dest, 3) == 0)
        {
            tmp = mpak->Header.flags >> 5;
            switch (tmp)
            {
                case TS_MESSAGE_OK:
                    fprintf(stdout, "OK");
                    break;
                case TS_MESSAGE_FROM_MAILBOX:
                    fprintf(stdout, "Delivered from mailbox");
                    break;                                                   
                case TS_MESSAGE_IN_MAILBOX:
                    fprintf(stdout, "Could not be delivered, stored in mailbox");
                    break;
                case TS_CANNOT_BE_REACHED:
                    fprintf(stdout, "Destination cannot be reached");
                    break;
                case TS_ILLEGAL_MESSAGE:
                    fprintf(stdout, "Bad message format");
                    break;
                case TS_NETWORK_CONGESTED:
                    fprintf(stdout, "Network congested");
                    break;
                case TS_TECHNICAL_ERROR:
                    fprintf(stdout, "Technical network error");
                    break;
                case TS_DESTINATION_BUSY:
                    fprintf(stdout, "Busy (not implemented)");
                    break;
                default: // we should never get here
                    fprintf(stdout, "Unknown");
            }
        }
        else
            fprintf(stdout, "N/A");    

        fprintf(stdout, "\nPacket Type: ");
        switch (mpak->Header.type)
        {
            case MPAK_TEXT:
                fprintf(stdout, "Text");
                break;
            case MPAK_DATA:
                fprintf(stdout, "Data");
                break;
            case MPAK_STATUS:
                fprintf(stdout, "Status");
                break;
            case MPAK_HPDATA:
                fprintf(stdout, "HP Data");
                break;
            default: // we should never get here
                fprintf(stdout, "Unknown");
        }

        // Use address list
        // if (mpak->Header.flags & SENDLIST_FLAG_MASK)
    
        sprintf(strtmp, "%.2X%.2X%.2X", mpak->Header.time[0], mpak->Header.time[1], mpak->Header.time[2]);
        tmp = strtoul(strtmp, NULL, 16);
        fprintf(stdout, "\nTime Stamp: ");
        if (tmp)
        {
            struct tm		time;
            unsigned char   tmstr[25];
		    int md, mt; // temp. intermediate values

            // Minutes since 1/1/1985, 12am, Mobitex Local Time
            time.tm_hour = (tmp % 1440) / 60;
            time.tm_min = (tmp % 1440) % 60;
            md = tmp / 1440;
            mt = (4291 + 10 * (md - (36525 * ((100 * md + 30690) / 36525)) / 100)) / 10;
            time.tm_year = 1983 + (100 * md + 30690) / 36525 + (mt / 249);
            time.tm_mon = (100 * mt) / 3061 - 1 - (12 * (mt / 429));
            time.tm_mday = mt - (((100 * mt) / 3061) * 3061) / 100;
            switch(time.tm_mon)
            {
                case 1:
                    strcpy(tmstr, "Jan");
                    break;
                case 2:
                    strcpy(tmstr, "Feb");
                    break;
                case 3:
                    strcpy(tmstr, "Mar");
                    break;
                case 4:
                    strcpy(tmstr, "Apr");
                    break;
                case 5:
                    strcpy(tmstr, "May");
                    break;
                case 6:
                    strcpy(tmstr, "Jun");
                    break;
                case 7:
                    strcpy(tmstr, "Jul");
                    break;
                case 8:
                    strcpy(tmstr, "Aug");
                    break;
                case 9:
                    strcpy(tmstr, "Sep");
                    break;
                case 10:
                    strcpy(tmstr, "Oct");
                    break;
                case 11:
                    strcpy(tmstr, "Nov");
                    break;
                case 12:
                    strcpy(tmstr, "Dec");
                    break;
                default: // we should never get here
                    strcpy(tmstr, "   ");
            }
            sprintf(tmstr+3, " %.2d %.2d:%.2d %.4d", time.tm_mday, time.tm_hour, time.tm_min, time.tm_year);
            fprintf(stdout, "%s", tmstr);
        }
        else
            fprintf(stdout, "N/A");

        fprintf(stdout, "\nPacket ID: %d", mpak->Header.packet);

        fprintf(stdout, "\n\nMobitex Packet (MPAK) Body");
        fprintf(stdout, "\n--------------------------");

        if(strmem(mpak->Data, "CMIME", sizeof(mpak->Data)))
            type = MIME;
        else if (strmem(mpak->Data, "CICAL", sizeof(mpak->Data)))
            type = ICAL;
        else
            type = RAW;

        switch(type) // parse into message structure (if known)
        {
            case MIME:
                //fprintf(stdout, "\nUnknown: 0x");
                for (s = 0; s < 9; ++s);
                    //fprintf(stdout, "%.2X", mpak->Data[s]);
        
                // header varies slightly if device is transmitting
                if (memcmp(mpak->Rosi.man, mpak->Header.dest, 3) == 0)
                {
                    tmp = mpak->Data[s];
                    strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                    strtmp[tmp] = '\0';
                    fprintf(stdout, "\nSource MAN: %s", strtmp);
                    s += (tmp + 2);
                }

                tmp = mpak->Data[s];
                strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                strtmp[tmp] = '\0';
                fprintf(stdout, "\nDestination MAN: %s", strtmp);
                s += (tmp + 2);

                //fprintf(stdout, "\nUnknown: 0x");
                for (tmp = 0; tmp < 5; ++tmp, ++s);
                    //fprintf(stdout, "%.2X", mpak->Data[s]);

                s += 6; // type (fixed)

                //fprintf(stdout, "\nUnknown: 0x");
                for (tmp = 0; tmp < 5; ++tmp, ++s);
                    //fprintf(stdout, "%.2X", mpak->Data[s]);

                switch (mpak->Data[s])
                {
                    case 0x54:
                        fprintf(stdout, "\nMessage Type: E-Mail Original (MIME)");
                        s += 1;

                        //fprintf(stdout, "\nUnknown: 0x");
                        for (tmp = 0; tmp < 10; ++tmp, ++s);
                        //fprintf(stdout, "%.2X", mpak->Data[s]);
                
                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nTo: %s", strtmp);
                        s += (tmp + 2);

                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nFrom: %s", strtmp);
                        s += (tmp + 1);

                        //fprintf(stdout, "\nUnknown: 0x");
                        for (tmp = 0; tmp < 7; ++tmp, ++s);
                        //fprintf(stdout, "%.2X", mpak->Data[s]);

                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nSubject: %s", strtmp);
                        s += (tmp + 1);

                        //fprintf(stdout, "\nUnknown: 0x");
                        for (tmp = 0; tmp < 15; ++tmp, ++s);
                        //fprintf(stdout, "%.2X", mpak->Data[s]);
                   
                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nBody: %s", strtmp);

                        break;
                    case 0x5C:
                    case 0x64:
                    case 0x6C:
                        fprintf(stdout, "\nMessage Type: E-Mail Forwarded (MIME)");
                        s += 1;
                        
                        //fprintf(stdout, "\nUnknown: 0x");
                        for (tmp = 0; tmp < 10; ++tmp, ++s);
                        //fprintf(stdout, "%.2X", mpak->Data[s]);
                
                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nTo: %s", strtmp);
                        s += (tmp + 2);

                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nFrom: %s", strtmp);
                        s += (tmp + 1);

                        //fprintf(stdout, "\nUnknown: 0x");
                        for (tmp = 0; tmp < 7; ++tmp, ++s);
                        //fprintf(stdout, "%.2X", mpak->Data[s]);

                        tmp = mpak->Data[s];
                        strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                        strtmp[tmp] = '\0';
                        fprintf(stdout, "\nSubject: %s", strtmp);
                        s += (tmp + 1);

                        //fprintf(stdout, "\nUnknown: 0x");
                        for (tmp = 0; tmp < 30; ++tmp, ++s);
                        //fprintf(stdout, "%.2X", mpak->Data[s]);
                
                        tmp = mpak->Data[s];
                        if (tmp == 0)
                            strcpy(strtmp, "N/A");
                        else
                        {
                            strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                            strtmp[tmp] = '\0';
                        }
                        fprintf(stdout, "\nBody: %s", strtmp);
                
                        break;

                    case 0x00:
                    default:
                        fprintf(stdout, "\nMessage Type: E-Mail Encrypted (MIME)\n");
                        s += 1;
                        hexdisplay(&(mpak->Data[s]), ((mpak->Rosi.blocks - 1) * 18) - s);
                }
                break;

            case ICAL:
                //fprintf(stdout, "\nUnknown: 0x");
                for (s = 0; s < 9; ++s);
                    //fprintf(stdout, "%.2X", mpak->Data[s]);
        
                tmp = mpak->Data[s];
                strncpy(strtmp, &(mpak->Data[s+1]), tmp);
                strtmp[tmp] = '\0';
                fprintf(stdout, "\nDestination MAN: %s", strtmp);
                s += (tmp + 2);

                //fprintf(stdout, "\nUnknown: 0x");
                for (tmp = 0; tmp < 5; ++tmp, ++s);
                    //fprintf(stdout, "%.2X", mpak->Data[s]);

                fprintf(stdout, "\nMessage Type: iCalendar (ICAL)\n");
                s += 6;
                hexdisplay(&(mpak->Data[s]), ((mpak->Rosi.blocks - 1) * 18) - s);
                break;

            case RAW: // just dump raw data, since the body format varies so much
            default:
                hexdisplay(mpak->Data, (mpak->Rosi.blocks - 1) * 18);
        }
    } // end if(!packet)
    else
        hexdisplay(mpak->Data, (mpak->Rosi.blocks - 1) * 18);
}
